import config.base

import os

class Configure(config.base.Configure):
  def __init__(self, framework):
    config.base.Configure.__init__(self, framework)
    self.headerPrefix = ''
    self.substPrefix  = ''
    self.argDB        = framework.argDB
    self.found        = 0
    self.compilers    = self.framework.require('config.compilers', self)
    self.libraries    = self.framework.require('config.libraries', self)
    return

  def configureHelp(self, help):
    help.addOption('BLAS', '-with-blas=<lib>', 'Use BLAS, optionally indicating the library')
    return

  def checkLib(self, fullLib):
    '''Checking for the BLAS library'''
    oldLibs              = self.framework.argDB['LIBS']
    (self.dir, self.lib) = os.path.split(fullLib)
    self.lib             = os.path.splitext(self.lib)[0]
    if self.lib.startswith('lib'): self.lib = self.lib[3:]
    if self.libraries.check(self.lib, 'ddot', libDir = self.dir, otherLibs = self.compilers.flibs, fortranMangle = 1):
      self.found = 1
    self.framework.argDB['LIBS'] = oldLibs
    return self.found

  def setOutput(self):
    if self.found:
      self.addSubstitution('BLAS_DIR', self.dir)
      libFlag = ''
      if self.dir: libFlag = '-L'+self.dir+' '
      libFlag += self.libraries.getLibArgument(self.lib)
      self.addSubstitution('BLAS_LIB', libFlag)
    return

  def configure(self):
    try:
      if self.framework.argDB.has_key('with-blas') and self.framework.argDB['with-blas']:
        fullLib = self.framework.argDB['with-blas']
      else:
        fullLib = 'libblas.a'
      if self.executeTest(self.checkLib, fullLib): return

      petscPath = os.path.abspath(os.path.join(self.argDB['PETSC_DIR'], '..', 'blaslapack', 'lib'))
      fullLib   = os.path.join(petscPath, 'libblas.a')
      if os.path.exists(fullLib) and self.executeTest(self.checkLib, fullLib): return

      petscPath = os.path.join(petscPath, 'lib'+self.argDB['BOPT'], self.argDB['PETSC_ARCH'])
      fullLib   = os.path.join(petscPath, 'libblas.a')
      if os.path.exists(fullLib) and self.executeTest(self.checkLib, fullLib): return
    finally:
      self.setOutput()
    return
